from distutils.log import debug
import math
from functools import partial
from cv2 import sqrt

import torch
import torch.nn as nn
import torch.nn.functional as F

import timm.models.vision_transformer
from timm.models.layers import DropPath, Mlp

# from lib.utils.misc import is_main_process
# from lib.models.mixformer.head import build_box_head
# from lib.utils.box_ops import box_xyxy_to_cxcywh, box_cxcywh_to_xyxy

from einops import rearrange
from itertools import repeat
import collections.abc
from ltr.external.PreciseRoIPooling.pytorch.prroi_pool import PrRoIPool2D
# from lib.models.mixformer_vit.score_pred_module import ScoreDecoder

# From PyTorch internals
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse

to_2tuple = _ntuple(2)

def box_cxcywh_to_xyxy(x):
    x_c, y_c, w, h = x.unbind(-1)
    b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
         (x_c + 0.5 * w), (y_c + 0.5 * h)]
    return torch.stack(b, dim=-1)

def box_xyxy_to_cxcywh(x):
    x0, y0, x1, y1 = x.unbind(-1)
    b = [(x0 + x1) / 2, (y0 + y1) / 2,
         (x1 - x0), (y1 - y0)]
    return torch.stack(b, dim=-1)


class PatchEmbed(nn.Module):
    """ 2D Image to Patch Embedding
    """
    def __init__(self, patch_size=16, in_chans=3, embed_dim=768, norm_layer=None, flatten=True):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        self.flatten = flatten

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        # B, C, H, W = x.shape
        x = self.proj(x)
        if self.flatten:
            x = x.flatten(2).transpose(1, 2).contiguous()  # BCHW -> BNC
        x = self.norm(x)
        return x

class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0., proj_drop=0.):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.qkv_mem = None

    def forward(self, x, t_h, t_w):
        """
        x is a concatenated vector of template and search region features.
        """
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple) 
        # b, head, n_q, c @ b, head, c, n_k => b, head, n_q, n_k
        # b, head, n_q, n_k @ b, head, n_k, c => b, head, n_q, c

        # q_mt, q_s = torch.split(q, [t_h * t_w * 2, s_h * s_w], dim=2)
        # k_mt, k_s = torch.split(k, [t_h * t_w * 2, s_h * s_w], dim=2)
        # v_mt, v_s = torch.split(v, [t_h * t_w * 2, s_h * s_w], dim=2)

        # # asymmetric mixed attention
        # attn = (q_mt @ k_mt.transpose(-2, -1)) * self.scale
        # attn = attn.softmax(dim=-1)
        # attn = self.attn_drop(attn)
        # x_mt = (attn @ v_mt).transpose(1, 2).reshape(B, t_h*t_w*2, C)

        # attn = (q_s @ k.transpose(-2, -1)) * self.scale
        # attn = attn.softmax(dim=-1)
        # attn = self.attn_drop(attn)
        # x_s = (attn @ v).transpose(1, 2).reshape(B, s_h*s_w, C)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, t_h * t_w, C)

        # x = torch.cat([x_mt, x_s], dim=1)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    def forward_self(self, x, t_h, t_w):
        return self.forward(x, t_h, t_w)
    
    def forward_cross(self, x, q_h, q_w):
        """
        x is a concatenated vector of template and search region features.
        """
        # B, N, C = x.shape
        # print("? ", q.shape)
        # x = torch.cat([kv, q], dim=1)

        # kv_N = kv.shape[1]
        q_N = q_h * q_w

        B, N, C = x.shape

        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple)
        
        q = q[:, :, :q_N] # b, head, n, c
        # k = k[:, :, q_N:]
        # v = v[:, :, q_N:]

        attn = (q @ k.transpose(-2, -1)) * self.scale # b, head, n_q, (n_kv + n_q)
        attn = attn.softmax(dim=-1)
        # attn = attn * (1.0 / kv_N)
        attn = self.attn_drop(attn) # b, h, n_q, c => b, n_q, h, c
        q = (attn @ v).transpose(1, 2).reshape(B, q_h * q_w, C)
        
        # x = torch.cat([x_mt, x_s], dim=1)
        q = self.proj(q)
        q = self.proj_drop(q)
        # print("! ", q.shape)
        return q

    # def forward_cross(self, kv, kv_h, kv_w, q, q_h, q_w):
    #     """
    #     x is a concatenated vector of template and search region features.
    #     """
    #     # B, N, C = x.shape
    #     # print("? ", q.shape)
    #     x = torch.cat([kv, q], dim=1)

    #     kv_N = kv.shape[1]
    #     q_N = q.shape[1]

    #     B, N, C = x.shape

    #     qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
    #     q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple)
        
    #     q = q[:, :, kv_N:] # b, head, n, c
    #     k = k[:, :, :kv_N]
    #     v = v[:, :, :kv_N]

    #     attn = (q @ k.transpose(-2, -1)) * self.scale # b, head, n_q, n_kv
    #     attn = attn.softmax(dim=-1)
    #     # attn = attn * (1.0 / kv_N)
    #     attn = self.attn_drop(attn) # b, h, n_q, c => b, n_q, h, c
    #     q = (attn @ v).transpose(1, 2).reshape(B, q_h * q_w, C)
        
    #     # x = torch.cat([x_mt, x_s], dim=1)
    #     q = self.proj(q)
    #     q = self.proj_drop(q)
    #     # print("! ", q.shape)
    #     return q


    def forward_test(self, x, s_h, s_w):
        """
        x is a concatenated vector of template and search region features.
        """
        B, N, C = x.shape
        qkv_s = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q_s, _, _ = qkv_s.unbind(0)   # make torchscript happy (cannot use tensor as tuple)
        qkv = torch.cat([self.qkv_mem, qkv_s], dim=3)
        _, k, v = qkv.unbind(0)

        attn = (q_s @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, s_h*s_w, C)

        x = self.proj(x)
        x = self.proj_drop(x)
        return x

    def set_online(self, x, t_h, t_w):
        """
        x is a concatenated vector of template and search region features.
        """
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        self.qkv_mem = qkv
        q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple) [B, num_heads, N, C//num_heads]

        # asymmetric mixed attention
        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)

        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):
    def __init__(
            self, dim, num_heads, mlp_ratio=4., qkv_bias=False, drop=0., attn_drop=0.,
            drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, cross_block=False):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.cross_block = cross_block
        if cross_block:
            self.norm_kv = norm_layer(dim)
            self.pooling_size = 8
            self.prpooling = PrRoIPool2D(pooled_height=self.pooling_size, pooled_width=self.pooling_size, spatial_scale=1.0/16)
            self.attn_cross = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)

        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path1 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)
        self.drop_path2 = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x, t_h, t_w):
        x = x + self.drop_path1(self.attn(self.norm1(x), t_h, t_w))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))
        return x
    
    def forward_self(self, x, t_h, t_w):
        assert self.cross_block == False
        x = x + self.drop_path1(self.attn(self.norm1(x), t_h, t_w))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))
        return x
    
    def forward_cross(self, kv, q, q_h, q_w, kv_crop_box):
        assert self.cross_block == True
        pass
        kv_h = q_h # assuming that search && template shares size
        kv_w = q_w
        B, C = kv.shape[:2]
        kv = kv.reshape(B, C, kv_h * kv_w).transpose(1,2) # b, l, c
        x = torch.cat([q, kv], dim=1)
        # kv = self.norm_kv(kv)
        # kv = kv.transpose(1,2).reshape(B, C, kv_h, kv_w)
        # ####
        # # TODO: add prroipooling for kv
        # # kv = self.prpooling(kv, kv_crop_box)
        # # kv = kv[:, :, 4:12, 4:12]
        # kv = kv.reshape(B,C,-1).transpose(1,2) # b, l, c
        # ####

        x = self.norm1(x)
        # cross_attn = self.attn_cross.forward_cross(kv, self.pooling_size, self.pooling_size, q, q_h, q_w)
        self_attn = self.attn.forward_cross(x, q_h, q_w)
        # q = q + (self.drop_path1(cross_attn) + self.drop_path1(self_attn)) * 0.5
        q = q + self.drop_path1(self_attn)
        # q = q + self.drop_path1(self.attn(self.norm1(q), q_h, q_w))
        q = q + self.drop_path2(self.mlp(self.norm2(q)))
        return q

    def forward_test(self, x, s_h, s_w):
        x = x + self.drop_path1(self.attn.forward_test(self.norm1(x), s_h, s_w))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))
        return x

    def set_online(self, x, t_h, t_w):
        x = x + self.drop_path1(self.attn.set_online(self.norm1(x), t_h, t_w))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))
        return x


class VisionTransformer(timm.models.vision_transformer.VisionTransformer):
    """ Vision Transformer with support for global average pooling
    """
    def __init__(self, img_size_s=256, img_size_t=128, patch_size=16, in_chans=3, num_classes=1000, embed_dim=768,
                 depth=12, num_heads=12, mlp_ratio=4., qkv_bias=True, drop_rate=0., attn_drop_rate=0.,
                 drop_path_rate=0., weight_init='', embed_layer=PatchEmbed, norm_layer=None, act_layer=None):
        super(VisionTransformer, self).__init__(img_size=224, patch_size=patch_size, in_chans=in_chans,
                                                num_classes=num_classes, embed_dim=embed_dim, depth=depth,
                                                num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                                                drop_rate=drop_rate, attn_drop_rate=attn_drop_rate,
                                                drop_path_rate=drop_path_rate, weight_init=weight_init,
                                                norm_layer=norm_layer, act_layer=act_layer)
        # self.is_template_branch = is_template_branch

        self.patch_embed = embed_layer(
            patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.Sequential(*[
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i],
                norm_layer=norm_layer, cross_block=((i%2) == 1)) for i in range(depth)])

        num_patches_s = (img_size_s // patch_size) ** 2
        num_patches_t = (img_size_t // patch_size) ** 2
        self.pos_embed_s = nn.Parameter(torch.zeros(1, num_patches_s, embed_dim))
        self.pos_embed_t = nn.Parameter(torch.zeros(1, num_patches_t, embed_dim))

        if weight_init != 'skip':
            self.init_weights(weight_init)

    def forward(self, x_t, output_layer=None, input_template_feats=None, kv_crop_box=None, is_template_branch=True):
        """
        :param x_t: (batch, c, 128, 128)
        :param x_s: (batch, c, 256, 256)
        :return:
        """
        x_t = self.patch_embed(x_t)  # BCHW-->BNC
        B, C = x_t.size(0), x_t.size(-1)
        H_t = W_t = int(math.sqrt(x_t.size(1)+0.1))

        x_t = x_t + self.pos_embed_t
        x = x_t

        x = self.pos_drop(x)

        output_feats = {}
        for id, blk in enumerate(self.blocks):
            if is_template_branch:
                x = blk(x, H_t, W_t)
            else:
                # forward for search image
                assert input_template_feats is not None
                assert kv_crop_box is not None
                if id in input_template_feats.keys():
                    # forward for cross blk
                    kv = input_template_feats[id]
                    x = blk.forward_cross(kv, x, H_t, W_t, kv_crop_box) 
                else:
                    # forward for self blk
                    x = blk(x, H_t, W_t)

            if output_layer is not None and id in output_layer:
                tmp_b, tmp_c = x.size(0), x.size(-1)
                tmp_h = tmp_w = int(math.sqrt(x.size(1)+0.1))
                x_reshape = x.transpose(1,2).reshape(tmp_b, tmp_c, tmp_h, tmp_w)
                output_feats[id] = x_reshape

        x_t = x

        x_t_2d = x_t.transpose(1, 2).reshape(B, C, H_t, W_t)
        if output_layer is None:
            return x_t_2d # , x_ot_2d, x_s_2d

        return x_t_2d, output_feats

    def forward_test(self, x):
        x = self.patch_embed(x)
        H_s = W_s = int(math.sqrt(x.size(1) + 0.1))

        x = x + self.pos_embed_s
        x = self.pos_drop(x)

        for blk in self.blocks:
            x = blk.forward_test(x, H_s, W_s)

        x = rearrange(x, 'b (h w) c -> b c h w', h=H_s, w=H_s)

        return self.template, x

    def set_online(self, x_t, x_ot):
        x_t = self.patch_embed(x_t)
        x_ot = self.patch_embed(x_ot)

        H_t = W_t = int(math.sqrt(x_t.size(1) + 0.1))
        # _, _, H_t, W_t = x_t.size()

        x_t = x_t + self.pos_embed_t
        x_ot = x_ot + self.pos_embed_t
        x_ot = x_ot.reshape(1, -1, x_ot.size(-1))  # [1, num_ot * H_t * W_t, C]
        x = torch.cat([x_t, x_ot], dim=1)

        x = self.pos_drop(x)

        for blk in self.blocks:
            x = blk.set_online(x, H_t, W_t)

        x_t = x[:, :H_t * W_t]
        x_ot = x[:, H_t * W_t:]
        x_t = rearrange(x_t, 'b (h w) c -> b c h w', h=H_t, w=W_t)
        x_ot = rearrange(x_ot.squeeze(0), '(b h w) c -> b c h w', h=H_t, w=W_t)

        self.template = x_t

        # return x_t, x_ot


def get_mixformer_vit(settings):
    img_size_s = settings.SEARCH_SIZE
    img_size_t = settings.TEMPLATE_SIZE
    if settings.VIT_TYPE == 'large_patch16':
        vit = VisionTransformer(
            img_size_s=img_size_s, img_size_t=img_size_t,
            patch_size=16, embed_dim=1024, depth=24, num_heads=16, mlp_ratio=4, qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6), drop_path_rate=0.1)
    elif settings.VIT_TYPE == 'base_patch16':
        vit = VisionTransformer(
            img_size_s=img_size_s, img_size_t=img_size_t,
            patch_size=16, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
            norm_layer=partial(nn.LayerNorm, eps=1e-6), drop_path_rate=0.1)
        
        # vit_template = VisionTransformer(
        #     img_size_s=img_size_s, img_size_t=img_size_t,
        #     patch_size=16, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
        #     norm_layer=partial(nn.LayerNorm, eps=1e-6), drop_path_rate=0.1, is_template_branch=True)
            
    else:
        raise KeyError(f"VIT_TYPE shoule set to 'large_patch16' or 'base_patch16'")

    if settings.BACKBONE_PRETRAINED:
        try:
            ckpt_path = settings.PRETRAINED_PATH
            ckpt = torch.load(ckpt_path, map_location='cpu')['model']
            # interpolate pos embedding of search and template
            new_dict = {}
            H_s = W_s = img_size_s // 16
            H_t = W_t = img_size_t // 16
            # print(ckpt.keys())
            # print([p[0] for p in vit.named_parameters()])

            for k, v in ckpt.items():
                if 'pos_embed' in k:
                    new_dict[k] = v
                    pos = ckpt[k]
                    C = pos.size(2)
                    pos_2d = pos[:, :-1, :].transpose(1,2).reshape(1, C, 14, 14)
                    pos_s_2d = F.interpolate(pos_2d, size=(H_s, W_s), mode='bilinear')
                    pos_s = pos_s_2d.flatten(2).transpose(1,2) # BNC
                    pos_t_2d = F.interpolate(pos_2d, size=(H_t, W_t), mode='bilinear')
                    pos_t = pos_t_2d.flatten(2).transpose(1,2) # BNC
                    pos_s_key = k.replace('pos_embed', 'pos_embed_s')
                    pos_t_key = k.replace('pos_embed', 'pos_embed_t')
                    new_dict[pos_s_key] = pos_s
                    new_dict[pos_t_key] = pos_t
                    print("Interpolating current pos_embed_s and pos_embed_t using {} done.".format(k))
                else:
                    new_dict[k] = v
            
            vit_param_names = [p[0] for p in vit.named_parameters()]
            for name in vit_param_names:
                if 'attn_cross' in name:
                    cpt_key = name.replace('attn_cross', 'attn')
                    new_dict[name] = new_dict[cpt_key].clone()
                if 'norm_kv' in name:
                    cpt_key = name.replace('norm_kv', 'norm1')
                    new_dict[name] = new_dict[cpt_key].clone()

            missing_keys, unexpected_keys = vit.load_state_dict(new_dict, strict=False)

        except:
            print("Warning: Pretrained ViT weights are not loaded")
    # return vit_template, vit_search
    return vit


class MixRoiVit(nn.Module):
    """ Mixformer tracking with score prediction module, whcih jointly perform feature extraction and interaction. """
    def __init__(self, backbone):
        """ Initializes the model.
        """
        super().__init__()
        # self.template_backbone = template_backbone
        # self.search_backbone = search_backbone
        self.backbone = backbone

    def forward(self, template, output_layers=None):
        # search: (b, c, h, w)
        if template.dim() == 5:
            template = template.squeeze(0)
        return self.backbone(template, output_layers)

    def forward_template(self, template, output_layers):
        if template.dim() == 5:
            template = template.squeeze(0)
        return self.backbone(template, output_layers, is_template_branch=True)
        # return self.template_backbone(template, output_layers)

    def forward_search(self, search, input_template_feats, kv_crop_box, output_layers):
        if search.dim() == 5:
            search = search.squeeze(0)

        return self.backbone(search, output_layers, input_template_feats, kv_crop_box, is_template_branch=False)
        # return self.search_backbone(search, output_layers)

    def forward_test(self, search, run_score_head=True, gt_bboxes=None):
        # search: (b, c, h, w)
        if search.dim() == 5:
            search = search.squeeze(0)
        template, search = self.backbone.forward_test(search)
        return template, search
        # search (b, 384, 20, 20)
        # Forward the corner head and score head
        # out, outputs_coord_new = self.forward_head(search, template, run_score_head, gt_bboxes)

        # return out, outputs_coord_new

    def set_online(self, template, online_template):
        if template.dim() == 5:
            template = template.squeeze(0)
        if online_template.dim() == 5:
            template = online_template.squeeze(0)
        self.backbone.set_online(template, online_template)


    # def forward_head(self, search, template, run_score_head=True, gt_bboxes=None):
    #     """
    #     :param search: (b, c, h, w), reg_mask: (b, h, w)
    #     :return:
    #     """
    #     out_dict = {}
    #     out_dict_box, outputs_coord = self.forward_box_head(search)
    #     out_dict.update(out_dict_box)
    #     if run_score_head:
    #         # forward the classification head
    #         if gt_bboxes is None:
    #             gt_bboxes = box_cxcywh_to_xyxy(outputs_coord.clone().view(-1, 4))
    #         # (b,c,h,w) --> (b,h,w)
    #         out_dict.update({'pred_scores': self.score_branch(search, template, gt_bboxes).view(-1)})

    #     return out_dict, outputs_coord


    # def forward_box_head(self, search):
    #     """
    #     :param search: (b, c, h, w)
    #     :return:
    #     """
    #     if self.head_type == "CORNER":
    #         # run the corner head
    #         b = search.size(0)
    #         outputs_coord = box_xyxy_to_cxcywh(self.box_head(search))
    #         outputs_coord_new = outputs_coord.view(b, 1, 4)
    #         out = {'pred_boxes': outputs_coord_new}
    #         return out, outputs_coord_new
    #     else:
    #         raise KeyError

def build_mixformer_vit_multi_score(settings=None, train=True):
    backbone = get_mixformer_vit(settings)  # backbone without positional encoding and attention mask
    # score_branch = ScoreDecoder(pool_size=4)  # the proposed score prediction module (SPM)
    # box_head = build_box_head(cfg)  # a simple corner head
    model = MixRoiVit(
        backbone,
        # box_head,
        # score_branch,
        # head_type=cfg.MODEL.HEAD_TYPE
    )

    # if cfg.MODEL.PRETRAINED_STAGE1 and train:
    #     try:
    #         ckpt_path = settings.stage1_model
    #         ckpt = torch.load(ckpt_path, map_location='cpu')
    #         missing_keys, unexpected_keys = model.load_state_dict(ckpt['net'], strict=False)
    #         # if is_main_process():
    #         #     print("missing keys:", missing_keys)
    #         #     print("unexpected keys:", unexpected_keys)
    #         #     print("Loading pretrained mixformer weights done.")
    #     except:
    #         print("Warning: Pretrained mixformer weights are not loaded")

    return model